import { Server, Player } from "@ns";

const CONSTANTS = {
    ServerBaseGrowthRate: 1.03, // Unadjusted Growth rate
    ServerMaxGrowthRate: 1.0035, // Maximum possible growth rate (max rate accounting for server security)
    ServerFortifyAmount: 0.002, // Amount by which server's security increases when its hacked/grown
    ServerWeakenAmount: 0.05, // Amount by which server's security decreases when weakened
}

const BitNodeMultipliers = {
    HackingLevelMultiplier: 1,

    ServerGrowthRate: 1,
    ServerMaxMoney: 1,
    ServerStartingMoney: 1,
    ServerStartingSecurity: 1,
    ServerWeakenRate: 1,

    HackExpGain: 1,

    ScriptHackMoney: 1,
}

export function calculateIntelligenceBonus(intelligence: number, weight = 1): number {
    return 1 + (weight * Math.pow(intelligence, 0.8)) / 600;
  }

export function calculateServerGrowth(server: Server, threads: number, p: Player, cores = 1): number {
    const numServerGrowthCycles = Math.max(Math.floor(threads), 0);

    //Get adjusted growth rate, which accounts for server security
    const growthRate = CONSTANTS.ServerBaseGrowthRate;
    let adjGrowthRate = 1 + (growthRate - 1) / server.hackDifficulty;
    if (adjGrowthRate > CONSTANTS.ServerMaxGrowthRate) {
        adjGrowthRate = CONSTANTS.ServerMaxGrowthRate;
    }

    //Calculate adjusted server growth rate based on parameters
    const serverGrowthPercentage = server.serverGrowth / 100;
    const numServerGrowthCyclesAdjusted =
        numServerGrowthCycles * serverGrowthPercentage * BitNodeMultipliers.ServerGrowthRate;

    //Apply serverGrowth for the calculated number of growth cycles
    const coreBonus = 1 + (cores - 1) / 16;
    return Math.pow(adjGrowthRate, numServerGrowthCyclesAdjusted * p.mults.hacking_grow * coreBonus);
}

/**
 * Returns the chance the player has to successfully hack a server
 */
export function calculateHackingChance(server: Server, player: Player): number {
    const hackFactor = 1.75;
    const difficultyMult = (100 - server.hackDifficulty) / 100;
    const skillMult = hackFactor * player.skills.hacking;
    const skillChance = (skillMult - server.requiredHackingSkill) / skillMult;
    const chance =
        skillChance *
        difficultyMult *
        player.mults.hacking_chance *
        calculateIntelligenceBonus(player.skills.intelligence, 1);
    if (chance > 1) {
        return 1;
    }
    if (chance < 0) {
        return 0;
    }

    return chance;
}

/**
 * Returns the amount of hacking experience the player will gain upon
 * successfully hacking a server
 */
export function calculateHackingExpGain(server: Server, player: Player): number {
    const baseExpGain = 3;
    const diffFactor = 0.3;
    if (server.baseDifficulty == null) {
        server.baseDifficulty = server.hackDifficulty;
    }
    let expGain = baseExpGain;
    expGain += server.baseDifficulty * diffFactor;

    return expGain * player.mults.hacking_exp * BitNodeMultipliers.HackExpGain;
}

/**
 * Returns the percentage of money that will be stolen from a server if
 * it is successfully hacked (returns the decimal form, not the actual percent value)
 */
export function calculatePercentMoneyHacked(server: Server, player: Player): number {
    // Adjust if needed for balancing. This is the divisor for the final calculation
    const balanceFactor = 240;

    const difficultyMult = (100 - server.hackDifficulty) / 100;
    const skillMult = (player.skills.hacking - (server.requiredHackingSkill - 1)) / player.skills.hacking;
    const percentMoneyHacked =
        (difficultyMult * skillMult * player.mults.hacking_money * BitNodeMultipliers.ScriptHackMoney) / balanceFactor;
    if (percentMoneyHacked < 0) {
        return 0;
    }
    if (percentMoneyHacked > 1) {
        return 1;
    }

    return percentMoneyHacked;
}

/**
 * Returns time it takes to complete a hack on a server, in seconds
 */
export function calculateHackingTime(server: Server, player: Player): number {
    const difficultyMult = server.requiredHackingSkill * server.hackDifficulty;

    const baseDiff = 500;
    const baseSkill = 50;
    const diffFactor = 2.5;
    let skillFactor = diffFactor * difficultyMult + baseDiff;
    // tslint:disable-next-line
    skillFactor /= player.skills.hacking + baseSkill;

    const hackTimeMultiplier = 5;
    const hackingTime =
        (hackTimeMultiplier * skillFactor) /
        (player.mults.hacking_speed * calculateIntelligenceBonus(player.skills.intelligence, 1));

    return hackingTime;
}

/**
 * Returns time it takes to complete a grow operation on a server, in seconds
 */
export function calculateGrowTime(server: Server, player: Player): number {
    const growTimeMultiplier = 3.2; // Relative to hacking time. 16/5 = 3.2

    return growTimeMultiplier * calculateHackingTime(server, player);
}

/**
 * Returns time it takes to complete a weaken operation on a server, in seconds
 */
export function calculateWeakenTime(server: Server, player: Player): number {
    const weakenTimeMultiplier = 4; // Relative to hacking time

    return weakenTimeMultiplier * calculateHackingTime(server, player);
}